@@ -1,8 +1,4 @@
-import functools
-import optparse
-import re
-import socket
-import sys
+import functools, optparse, re, socket, sys
 
 from django.core.management import base as management_base
 from django.core.management.commands import runserver
@@ -111,7 +107,7 @@ def make_location(loc_dict, stores=None, servername=None):
     else:
         raise management_base.CommandError('Invalid location type "%s".' % (loc_type,))
 
-    url = loc_conf.pop('url', loc_conf.pop('prefix', '')+'(.+)')
+    url = loc_conf.pop('url', loc_conf.pop('prefix', '') + '(.+)')
     store_id = loc_conf.pop('store')
     kwargs = {
         'registry': stores[store_id]['registry'],
@@ -126,6 +122,8 @@ class Command(management_base.BaseCommand):
             help='Tells Django to use an IPv6 address.'),
         optparse.make_option('--allrequests', action='store_true', dest='allrequests', default=False,
             help='Process also non-push requests.'),
+        optparse.make_option('--forcehost', action='store', type='string', dest='forcehost',
+            help='When processing also non-push requests, force all request to one host.'),
     )
     help = "Starts a push server."
     args = '[optional port number, or ipaddr:port]'
@@ -136,6 +134,9 @@ class Command(management_base.BaseCommand):
     def handle(self, addrport='', *args, **options):
         self.use_ipv6 = options.get('use_ipv6')
         self.allrequests = options.get('allrequests')
+        self.forcehost = options.get('forcehost')
+        if self.forcehost and not self.allrequests:
+            raise management_base.CommandError("'forcehost' can be set only with 'allrequests'.")
         self._raw_ipv6 = False
         if args:
             raise management_base.CommandError('Usage is runpushserver %s' % self.args)
@@ -205,22 +206,49 @@ class Command(management_base.BaseCommand):
         conf['locations'] = map(functools.partial(make_location, stores=conf['store'], servername=conf['servername']), conf['locations'])
 
         if self.allrequests:
-            class FallbackHandlerWithServerName(web.FallbackHandler):
-                def set_default_headers(self):
-                    if conf.get('servername', None):
-                        # TODO: Does not really work, https://github.com/sunblaze-ucb/threader/issues/4
-                        self.set_header('Server', conf.get('servername', None))
+            class SetServenameMiddleware(object):
+                def __init__(self, application):
+                    self.application = application
 
-            wsgi_app = tornado_wsgi.WSGIContainer(django_wsgi.get_wsgi_application())
+                def __call__(self, environ, start_response):
+                    def servername_start_response(status, response_headers, exc_info=None):
+                        if conf.get('servername', None):
+                            header_set = set(k.lower() for (k, v) in response_headers)
+                            if 'server' not in header_set:
+                                response_headers.append(('Server', conf.get('servername', None)))
+                        return start_response(status, response_headers, exc_info)
+
+                    return self.application(environ, servername_start_response)
+
+            wsgi_app = tornado_wsgi.WSGIContainer(SetServenameMiddleware(django_wsgi.get_wsgi_application()))
             conf['locations'] += (
-                ('.*', FallbackHandlerWithServerName, {'fallback': wsgi_app}),
+                ('.*', web.FallbackHandler, {'fallback': wsgi_app}),
             )
 
         import logging
         logging.getLogger().setLevel('INFO')
         tornado_options.enable_pretty_logging()
 
-        httpserver.HTTPServer(web.Application(conf['locations'])).listen(conf['port'], conf['address'])
+        if self.forcehost:
+            class RedirectHandler(web.RequestHandler):
+                def initialize(self, forcehost):
+                    self._forcehost = forcehost
+
+                def prepare(self):
+                    url = self.request.protocol + '://' + self._forcehost + self.request.uri
+                    self.redirect(url, permanent=True)
+
+                def set_default_headers(self):
+                    if conf.get('servername', None):
+                        self.set_header('Server', conf.get('servername', None))
+
+            application = web.Application()
+            application.add_handlers('^%s$' % re.escape(self.forcehost.lower().split(':')[0]), conf['locations'])
+            application.add_handlers('.*$', (('.*', RedirectHandler, {'forcehost': self.forcehost}),))
+        else:
+            application = web.Application(conf['locations'])
+
+        httpserver.HTTPServer(application).listen(conf['port'], conf['address'])
 
         try:
             ioloop.IOLoop.instance().start()
